(ns com.aurellem.exp.item-bridge
  (:use (com.aurellem.gb saves util constants gb-driver vbm items assembly))
  (:use (com.aurellem.run title save-corruption))
  ;;(:use (com.aurellem.exp pokemon))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(defn corrupt-item-state []
  (second (destroy-item-end-of-list-marker)))

(defn corrupt-item-state []
  (read-state "corrupt-items"))

(defn view-memory-range [state start end]
  (dorun
   (map (fn [loc val]
          (println (format "%04X : %02X" loc val)))
        (range start end) (subvec (vec (memory state)) start end)))
  state)

(defn almost-broken
  "if one more memory location is turned into 0x03, the game crashes."
  [n]
  (view-memory-range
   (set-inv-mem (mid-game)
                (concat [0xFF] (repeat 64 0x03)
                        (subvec (vec (memory (mid-game)))
                                (+ item-list-start 65)
                                (+ item-list-start 65 n))
                        (repeat (- 255 65 n) 0x03)))
   item-list-start (+ item-list-start 255)))

(defn actually-broken
  "if this memory location is turned into 0x03, the game crashes."
  []
  (set-memory (mid-game) 0xD35D 0x03))


;; (almost-broken 20) more or less works

(defn capture-program-counter
  "records the program counter for each tick"
  [^SaveState state ticks]
  (let [i (atom 0)]
    (reduce (fn [[program-counters state] _]
              (println (swap! i inc))
               [(conj program-counters (PC state))
                (tick state)])
              [[] state]
              (range ticks))))


(defn capture-program-counter
  [^SaveState state ticks]
  (tick state)
  
  (loop [i 0
         pcs []]
    (if (= i ticks)
      (filter (partial < 0x2000)(sort (set pcs)))
      (do 
        (com.aurellem.gb.Gb/tick)
        (recur (inc i)
               (conj pcs (first (registers))))))))

(defn loop-program []
  [0x00     ;0xD31D  ;; disable-interrupts

   0xC3     ;; loop forever
   0x1D
   0xD3])

(def map-function-address-start 0xD36D)

(defn test-loop []
  (continue!
   (-> (mid-game) 
       (set-memory-range 0xD31D (loop-program))
       (set-memory-range
        map-function-address-start
        [0xD3 0x1D]))))
       
(defn-memo corrupt-moves []
  (concat
   (first
    (->>
     [[] (mid-game)]
     (advance [:b] [:b :start])
     (advance [] [:d])
     (play-moves [[] [] [] [:d] [] [] [] [:d] [] [] [:a]])
     scroll-text
     (play-moves
      ;; this section is copied from speedrun-2942
      ;; and corrupts the save so that the end-of-list marker
      ;; for the pokemon roster is destroyed, but the save is still
      ;; playable.
      [[] [] [] [] [] [] [] [] [] [] [] [] [] [] [] [] [] [] [] []
       [] [] [] [] [] [] [] [] [] [] [:select] [:restart]])
     (play-moves
      (first (title)))
     (advance [] [:start])
     (advance [] [:a])
     (advance [:a] [:a :start])))
   [[]]))

(defn corrupt
  "enter the codes to destroy the
   pokemon list using save corruption"
  ([^SaveState state]
  (run-moves
   state
   (corrupt-moves)))
  ([] (corrupt @current-state)))

(defn mid-game-corrupt []
  (read-state "corrupt-mid-game"))

(defn gen-start-game-corrupt []
  (->> (finish-title)
       (advance [:b] [:a :b :start])
       (play-moves (corrupt-moves))))

(defn start-game-corrupt []
  (read-state "corrupt-start-game"))

(defn test-memory-fun [n]
  (capture-program-counter
   (set-memory-range
    (tick (mid-game))
    0xD36D
    [0 0])
   n))

;;(def good (test-memory-fun 17000))

;;(def bad (test-memory-fun 18000))



(defn menu-open-state []
  (read-state "menu-open"))

(defn prepare-memory
  ([^SaveState state]
     (-> state
         (set-memory-range 0xD31D (loop-program))
         (set-memory-range 0xD36D [0x1D 0xD3])))
  ([] (prepare-memory @current-state)))

(def memory-function-address-start 0xD36D)

(defn read-map-function-address
  ([^SaveState state]
     (let [mem (memory state)]
       [(aget mem memory-function-address-start)
        (aget mem (inc memory-function-address-start))]))
  ([] (read-map-function-address @current-state)))

(defn succesful-PC-capture
  "This function demonstrates successful PC capturing by
   setting 0xD36D to the value of the start location of
   a specially prepared program.

   You must run the function and then exit the open menu
   to see the effect."
  []
  (dorun
   (map #(println (Integer/toHexString %))
        (capture-program-counter 
         (prepare-memory (menu-open-state))
         9000000))))

(defn trampoline-assembly [^SaveState state]
  (flatten
   [0x3E ;; 
    0x3E ;; load lemonade into A
    
    0xEA
    0x1D
    0xD3 ;; set first item to lemonade
    
    0xC3 ;; return control to the game via absolute junp.
    (read-map-function-address state)
    ]))

(defn test-trampoline
  "Demonstrates item-program execution via the map-function that
   returns control to the main pokemon game after one loop."
  []
  (let [insertion-address 0xD33D
        insertion-address-bits [0x3D 0xD3]]
  (->
   (menu-open-state)
   (set-memory-range
    insertion-address
    (trampoline-assembly (menu-open-state)))
   (set-memory-range 
    memory-function-address-start
    insertion-address-bits))))



;;map-functions

;; Room, Upstairs        0xBO 0x40
;; Room, Downstairs      0x02 0x41
;; Pallet Town           0xFA 0x4D
;; Rival's House         0x3B 0x5C
;; Oak's Lab             0x92 0x43
;; Route 1               0x58 0x43
;; Viridian City         0x2A 0x50
;; Viridian PkmnCenter   0x5D 0x42
;; Viridian Mart         0x7A 0x4C
;; Viridian House #1     0x79 0x4D
;; Viridian House #2     0xD3 0x4D
;; Route 22              0xB3 0x4E
;; League Front Gate     0x8D 0x5F